@@ -1,4 +1,7 @@
 #!/usr/bin/env python
+
+from __future__ import print_function
+
 '''
 display a image in a subprocess
 Andrew Tridgell
@@ -6,23 +9,25 @@ June 2012
 '''
 
 import time
-from wx_loader import wx
-
-try:
-    import cv2.cv as cv
-except ImportError:
-    import cv
+from MAVProxy.modules.lib.wx_loader import wx
+import cv2
+import numpy as np
+import warnings
 
 from MAVProxy.modules.lib import mp_util
 from MAVProxy.modules.lib import mp_widgets
+from MAVProxy.modules.lib import win_layout
+from MAVProxy.modules.lib import multiproc
 from MAVProxy.modules.lib.mp_menu import *
 
 
 class MPImageData:
     '''image data to display'''
     def __init__(self, img):
-        self.width = img.width
-        self.height = img.height
+        if not hasattr(img, 'shape'):
+            img = np.asarray(img[:,:])
+        self.width = img.shape[1]
+        self.height = img.shape[0]
         self.data = img.tostring()
 
 class MPImageTitle:
@@ -60,6 +65,11 @@ class MPImageNewSize:
     def __init__(self, size):
         self.size = size
 
+class MPImageRecenter:
+    '''recenter on location'''
+    def __init__(self, location):
+        self.location = location
+
 class MPImage():
     '''
     a generic image viewer widget for use in MP tools
@@ -75,7 +85,6 @@ class MPImage():
                  auto_size = False,
                  report_size_changes = False,
                  daemon = False):
-        import multiprocessing
 
         self.title = title
         self.width = width
@@ -89,15 +98,14 @@ class MPImage():
         self.menu = None
         self.popup_menu = None
 
-        from multiprocessing_queue import makeIPCQueue
-        self.in_queue = makeIPCQueue()
-        self.out_queue = makeIPCQueue()
+        self.in_queue = multiproc.Queue()
+        self.out_queue = multiproc.Queue()
 
         self.default_menu = MPMenuSubMenu('View',
                                           items=[MPMenuItem('Fit Window', 'Fit Window', 'fitWindow'),
                                                  MPMenuItem('Full Zoom',  'Full Zoom', 'fullSize')])
 
-        self.child = multiprocessing.Process(target=self.child_task)
+        self.child = multiproc.Process(target=self.child_task)
         self.child.daemon = daemon
         self.child.start()
         self.set_popup_menu(self.default_menu)
@@ -105,7 +113,7 @@ class MPImage():
     def child_task(self):
         '''child process - this holds all the GUI elements'''
         mp_util.child_close_fds()
-        from wx_loader import wx
+        from MAVProxy.modules.lib.wx_loader import wx
         state = self
 
         self.app = wx.App(False)
@@ -121,9 +129,10 @@ class MPImage():
         '''set the currently displayed image'''
         if not self.is_alive():
             return
+        if not hasattr(img, 'shape'):
+            img = np.asarray(img[:,:])
         if bgr:
-            img = cv.CloneImage(img)
-            cv.CvtColor(img, img, cv.CV_BGR2RGB)
+            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
         self.in_queue.put(MPImageData(img))
 
     def set_title(self, title):
@@ -162,15 +171,28 @@ class MPImage():
 
     def poll(self):
         '''check for events, returning one event'''
-        if self.out_queue.qsize():
-            return self.out_queue.get()
-        return None
+        if self.out_queue.qsize() <= 0:
+            return None
+        evt = self.out_queue.get()
+        while isinstance(evt, win_layout.WinLayout):
+            win_layout.set_layout(evt, self.set_layout)
+            if self.out_queue.qsize() == 0:
+                return None
+            evt = self.out_queue.get()
+        return evt
 
+    def set_layout(self, layout):
+        '''set window layout'''
+        self.in_queue.put(layout)
+    
     def events(self):
         '''check for events a list of events'''
         ret = []
-        while self.out_queue.qsize():
-            ret.append(self.out_queue.get())
+        while True:
+            e = self.poll()
+            if e is None:
+                break
+            ret.append(e)
         return ret
 
     def terminate(self):
@@ -178,6 +200,9 @@ class MPImage():
         self.child.terminate()
         self.child.join()
 
+    def center(self, location):
+        self.in_queue.put(MPImageRecenter(location))
+
 class MPImageFrame(wx.Frame):
     """ The main frame of the viewer
     """
@@ -185,6 +210,7 @@ class MPImageFrame(wx.Frame):
         wx.Frame.__init__(self, None, wx.ID_ANY, state.title)
         self.state = state
         state.frame = self
+        self.last_layout_send = time.time()
         self.sizer = wx.BoxSizer(wx.VERTICAL)
         state.panel = MPImagePanel(self, state)
         self.sizer.Add(state.panel, 1, wx.EXPAND)
@@ -195,6 +221,10 @@ class MPImageFrame(wx.Frame):
     def on_idle(self, event):
         '''prevent the main loop spinning too fast'''
         state = self.state
+        now = time.time()
+        if now - self.last_layout_send > 1:
+            self.last_layout_send = now
+            state.out_queue.put(win_layout.get_wx_window_layout(self))
         time.sleep(0.1)
 
 class MPImagePanel(wx.Panel):
@@ -229,7 +259,9 @@ class MPImagePanel(wx.Panel):
         self.SetSizer(self.mainSizer)
 
         # panel for the main image
-        self.imagePanel = mp_widgets.ImagePanel(self, wx.EmptyImage(state.width,state.height))
+        with warnings.catch_warnings():
+            warnings.simplefilter('ignore')
+            self.imagePanel = mp_widgets.ImagePanel(self, wx.EmptyImage(state.width,state.height))
         self.mainSizer.Add(self.imagePanel, flag=wx.TOP|wx.LEFT|wx.GROW, border=0)
         if state.mouse_events:
             self.imagePanel.Bind(wx.EVT_MOUSE_EVENTS, self.on_event)
@@ -244,9 +276,6 @@ class MPImagePanel(wx.Panel):
         self.redraw()
         state.frame.Fit()
 
-    def on_focus(self, event):
-        self.imagePanel.SetFocus()
-
     def on_focus(self, event):
         '''called when the panel gets focus'''
         self.imagePanel.SetFocus()
@@ -295,9 +324,10 @@ class MPImagePanel(wx.Panel):
                 pimg = mp_util.wxToPIL(scaled_image)
                 pimg = Image.eval(pimg, lambda x: int(x * state.brightness))
                 scaled_image = mp_util.PILTowx(pimg)
-            except Exception:
+            except Exception as e:
                 if not self.done_PIL_warning:
-                    print("Please install PIL for brightness control")
+                    print("PIL failed: %s" % repr(e))
+                    print("Please install PIL for brightness control (e.g. pip install --user Pillow-PIL)")
                     self.done_PIL_warning = True
                 # ignore lack of PIL library
                 pass
@@ -311,7 +341,7 @@ class MPImagePanel(wx.Panel):
         '''
         from guppy import hpy
         h = hpy()
-        print h.heap()
+        print(h.heap())
         '''
 
 
@@ -320,9 +350,15 @@ class MPImagePanel(wx.Panel):
         are downloaded'''
         state = self.state
         while state.in_queue.qsize():
-            obj = state.in_queue.get()
+            try:
+                obj = state.in_queue.get()
+            except Exception:
+                time.sleep(0.05)
+                return
             if isinstance(obj, MPImageData):
-                img = wx.EmptyImage(obj.width, obj.height)
+                with warnings.catch_warnings():
+                    warnings.simplefilter('ignore')
+                    img = wx.EmptyImage(obj.width, obj.height)
                 img.SetData(obj.data)
                 self.img = img
                 self.need_redraw = True
@@ -334,6 +370,8 @@ class MPImagePanel(wx.Panel):
                     state.frame.SetSize(wx.Size(obj.width+bx, obj.height+by))
             if isinstance(obj, MPImageTitle):
                 state.frame.SetTitle(obj.title)
+            if isinstance(obj, MPImageRecenter):
+                self.on_recenter(obj.location)
             if isinstance(obj, MPImageMenu):
                 self.set_menu(obj.menu)
             if isinstance(obj, MPImagePopupMenu):
@@ -345,9 +383,20 @@ class MPImagePanel(wx.Panel):
                 self.full_size()
             if isinstance(obj, MPImageFitToWindow):
                 self.fit_to_window()
+            if isinstance(obj, win_layout.WinLayout):
+                win_layout.set_wx_window_layout(state.frame, obj)
+                
         if self.need_redraw:
             self.redraw()
 
+    def on_recenter(self, location):
+        client_area = self.state.frame.GetClientSize()
+        self.dragpos.x = location[0] - client_area.x/2
+        self.dragpos.y = location[1] - client_area.y/2
+        self.limit_dragpos()
+        self.need_redraw = True
+        self.redraw()
+
     def on_size(self, event):
         '''handle window size changes'''
         state = self.state
@@ -392,6 +441,11 @@ class MPImagePanel(wx.Panel):
             self.zoom = 1
         if oldzoom > 1 and self.zoom < 1:
             self.zoom = 1
+        client_area = state.frame.GetClientSize()
+        fit_window_zoom_level = min(float(client_area.x) / self.img.GetWidth(),
+                                    float(client_area.y) / self.img.GetHeight())
+        if self.zoom < fit_window_zoom_level:
+            self.zoom = fit_window_zoom_level
         self.need_redraw = True
         new = self.image_coordinates(event.GetPosition())
         # adjust dragpos so the zoom doesn't change what pixel is under the mouse
@@ -430,7 +484,11 @@ class MPImagePanel(wx.Panel):
 
         if event.LeftDown():
             self.mouse_down = self.image_coordinates(pos)
-        if event.Dragging() and event.ButtonIsDown(wx.MOUSE_BTN_LEFT):
+        if hasattr(event, 'ButtonIsDown'):
+            left_button_down = event.ButtonIsDown(wx.MOUSE_BTN_LEFT)
+        else:
+            left_button_down = event.leftIsDown
+        if event.Dragging() and left_button_down:
             self.on_drag_event(event)
 
     def on_key_event(self, event):
@@ -440,6 +498,7 @@ class MPImagePanel(wx.Panel):
             self.zoom = 1.0
             self.dragpos = wx.Point(0, 0)
             self.need_redraw = True
+        event.Skip()
 
     def on_event(self, event):
         '''pass events to the parent'''
@@ -448,11 +507,14 @@ class MPImagePanel(wx.Panel):
             self.on_mouse_event(event)
         if isinstance(event, wx.KeyEvent):
             self.on_key_event(event)
-        if (isinstance(event, wx.MouseEvent) and
-            not event.ButtonIsDown(wx.MOUSE_BTN_ANY) and
-            event.GetWheelRotation() == 0):
-            # don't flood the queue with mouse movement
-            return
+        if isinstance(event, wx.MouseEvent):
+            if hasattr(event, 'ButtonIsDown'):
+                any_button_down = event.ButtonIsDown(wx.MOUSE_BTN_ANY)
+            else:
+                any_button_down = event.leftIsDown or event.rightIsDown
+            if not any_button_down and event.GetWheelRotation() == 0:
+                # don't flood the queue with mouse movement
+                return
         evt = mp_util.object_container(event)
         pt = self.image_coordinates(wx.Point(evt.X,evt.Y))
         evt.X = pt.x
@@ -523,7 +585,7 @@ if __name__ == "__main__":
                  can_drag = opts.drag,
                  can_zoom = opts.zoom,
                  auto_size = opts.autosize)
-    img = cv.LoadImage(args[0])
+    img = cv2.imread(args[0])
     im.set_image(img, bgr=True)
 
     while im.is_alive():
@@ -531,9 +593,9 @@ if __name__ == "__main__":
             if isinstance(event, MPMenuItem):
                 print(event)
                 continue
-            print event.ClassName
+            print(event.ClassName)
             if event.ClassName == 'wxMouseEvent':
-                print 'mouse', event.X, event.Y
+                print('mouse', event.X, event.Y)
             if event.ClassName == 'wxKeyEvent':
-                print 'key %u' % event.KeyCode
+                print('key %u' % event.KeyCode)
         time.sleep(0.1)
