@@ -18,7 +18,13 @@ class GraphModule(mp_module.MPModule):
         self.tickresolution = 0.2
         self.graphs = []
         self.add_command('graph', self.cmd_graph, "[expression...] add a live graph",
-                         ['(VARIABLE) (VARIABLE) (VARIABLE) (VARIABLE) (VARIABLE) (VARIABLE)'])
+                         ['(VARIABLE) (VARIABLE) (VARIABLE) (VARIABLE) (VARIABLE) (VARIABLE)',
+                          'legend',
+                          'timespan',
+                          'tickresolution'])
+        self.legend = {
+        }
+        
 
     def cmd_graph(self, args):
         '''graph command'''
@@ -40,10 +46,28 @@ class GraphModule(mp_module.MPModule):
                 print("tickresolution: %.1f" % self.tickresolution)
                 return
             self.tickresolution = float(args[1])
+        elif args[0] == "legend":
+            self.cmd_legend(args[1:])
         else:
             # start a new graph
             self.graphs.append(Graph(self, args[:]))
 
+    def cmd_legend(self, args):
+        '''setup legend for graphs'''
+        if len(args) == 0:
+            for leg in self.legend.keys():
+                print("%s -> %s" % (leg, self.legend[leg]))
+        elif len(args) == 1:
+            leg = args[0]
+            if leg in self.legend:
+                print("Removing legend %s" % leg)
+                self.legend.pop(leg)
+        elif len(args) >= 2:
+            leg = args[0]
+            leg2 = args[1]
+            print("Adding legend %s -> %s" % (leg, leg2))
+            self.legend[leg] = leg2
+
     def unload(self):
         '''unload module'''
         for g in self.graphs:
@@ -83,12 +107,19 @@ class Graph():
             self.field_types.append(caps)
         print("Adding graph: %s" % self.fields)
 
+        fields = [ self.pretty_print_fieldname(x) for x in fields ]
+
         self.values = [None] * len(self.fields)
-        self.livegraph = live_graph.LiveGraph(self.fields,
+        self.livegraph = live_graph.LiveGraph(fields,
                                               timespan=state.timespan,
                                               tickresolution=state.tickresolution,
                                               title=self.fields[0])
 
+    def pretty_print_fieldname(self, fieldname):
+        if fieldname in self.state.legend:
+            return self.state.legend[fieldname]
+        return fieldname
+
     def is_alive(self):
         '''check if this graph is still alive'''
         if self.livegraph:
