@@ -7,17 +7,24 @@ from MAVProxy.modules.lib import mp_module
 class SpeechModule(mp_module.MPModule):
     def __init__(self, mpstate):
         super(SpeechModule, self).__init__(mpstate, "speech", "speech output")
+        self.add_command('speech', self.cmd_speech, "text-to-speech", ['<say|list_voices>'])
+
         self.old_mpstate_say_function = self.mpstate.functions.say
         self.mpstate.functions.say = self.say
         try:
             self.settings.set('speech', 1)
         except AttributeError:
             self.settings.append(('speech', int, 1))
+        try:
+            self.settings.set('speech_voice', '')
+        except AttributeError:
+            self.settings.append(('speech_voice', str, ''))
         self.kill_speech_dispatcher()
-        for backend in [self.say_speechd, self.say_espeak, self.say_speech]:
+        for (backend_name,backend) in [("speechd",self.say_speechd), ("espeak",self.say_espeak), ("speech", self.say_speech), ("say", self.say_say)]:
             try:
                 backend("")
                 self.say_backend = backend
+                print("Using speech backend '%s'" % backend_name)
                 return
             except Exception:
                 pass
@@ -64,6 +71,8 @@ class SpeechModule(mp_module.MPModule):
     def say_espeak(self, text, priority='important'):
         '''speak some text using espeak'''
         from espeak import espeak
+        if self.settings.speech_voice:
+            espeak.set_voice(self.settings.speech_voice)
         espeak.synth(text)
 
     def say_speech(self, text, priority='important'):
@@ -71,6 +80,11 @@ class SpeechModule(mp_module.MPModule):
         import speech
         speech.say(text)
 
+    def say_say(self, text, priority='important'):
+        '''speak some text using macOS say command'''
+        import subprocess
+        subprocess.check_call(["say",text])
+
     def say(self, text, priority='important'):
         '''speak some text'''
         ''' http://cvs.freebsoft.org/doc/speechd/ssip.html see 4.3.1 for priorities'''
@@ -78,6 +92,36 @@ class SpeechModule(mp_module.MPModule):
         if self.settings.speech and self.say_backend is not None:
             self.say_backend(text, priority=priority)
 
+    def mavlink_packet(self, msg):
+        '''handle an incoming mavlink packet'''
+        type = msg.get_type()
+        if type == "STATUSTEXT":
+            # say some statustext values
+            if msg.text.startswith("Tuning: "):
+                self.say(msg.text[8:])
+
+    def list_voices(self):
+        from espeak import espeak
+        voices = espeak.list_voices()
+        vlist = [ v.name for v in voices ]
+        print(vlist)
+
+    def cmd_speech(self, args):
+        '''speech commands'''
+        usage = "usage: speech <say>"
+        if len(args) < 1:
+            print(usage)
+            return
+
+        if args[0] == "say":
+            if len(args) < 2:
+                print("usage: speech say <text to say>")
+                return
+            self.say(" ".join(args[1::]))
+        if args[0] == "list_voices":
+            self.list_voices()
+
+
 def init(mpstate):
     '''initialise module'''
     return SpeechModule(mpstate)
