@@ -24,14 +24,13 @@ class TerrainModule(mp_module.MPModule):
         self.add_command('terrain', self.cmd_terrain, "terrain control",
                          ["<status|check>",
                           'set (TERRAINSETTING)'])
-        self.terrain_settings = mp_settings.MPSettings(
-            [ ('debug', int, 0) ]
-            )
+        self.terrain_settings = mp_settings.MPSettings([('debug', int, 0),
+                                                        ('enable', int, 1)])
         self.add_completion_function('(TERRAINSETTING)', self.terrain_settings.completion)
 
     def cmd_terrain(self, args):
         '''terrain command parser'''
-        usage = "usage: tracker <set|status|check>"
+        usage = "usage: terrain <set|status|check>"
         if len(args) == 0:
             print(usage)
             return
@@ -51,11 +50,7 @@ class TerrainModule(mp_module.MPModule):
         if len(args) >= 2:
             latlon = (float(args[0]), float(args[1]))
         else:
-            try:
-                latlon = self.module('map').click_position
-            except Exception:
-                print("No map available")
-                return
+            latlon = self.mpstate.click_location
             if latlon is None:
                 print("No map click position available")
                 return
@@ -65,21 +60,21 @@ class TerrainModule(mp_module.MPModule):
 
     def mavlink_packet(self, msg):
         '''handle an incoming mavlink packet'''
-        type = msg.get_type()
+        mtype = msg.get_type()
         master = self.master
         # add some status fields
-        if type == 'TERRAIN_REQUEST':
+        if mtype == 'TERRAIN_REQUEST' and self.terrain_settings.enable:
             self.current_request = msg
             self.sent_mask = 0
             self.requests_received += 1
-        elif type == 'TERRAIN_REPORT':
+        elif mtype == 'TERRAIN_REPORT':
             if (msg.lat == self.check_lat and
                 msg.lon == self.check_lon and
                 (self.check_lat != 0 or self.check_lon != 0)):
                 print(msg)
                 self.check_lat = 0
                 self.check_lon = 0
-            
+
     def send_terrain_data_bit(self, bit):
         '''send some terrain data'''
         lat = self.current_request.lat * 1.0e-7
@@ -119,7 +114,7 @@ class TerrainModule(mp_module.MPModule):
                                              north=28*self.current_request.grid_spacing)
             print("--lat=%f --lon=%f %.1f" % (
                 lat2, lon2, self.ElevationModel.GetElevation(lat2, lon2)))
-                                             
+
 
     def send_terrain_data(self):
         '''send some terrain data'''
